# distutils: language = c
# cython: language_level=3
# cython: linetrace=True
# cython: infer_types=True

import numpy as np
cimport numpy as np
cimport cython

DTYPE = int
ctypedef np.int_t DTYPE_t
DTYPE8 = np.int8
ctypedef np.int8_t DTYPE8_t

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef idx_to_signal(DTYPE_t[:] idx, int step):
    """Converts an index array to a signal array.

    Parameters
    ----------
    idx : array_like
        Index array.
    step : int
        Step size.

    Returns
    -------
    signal : ndarray
        Signal array.

    Notes
    -----
    This method uses Cython to speed up the computation.

    Benchmark for 1,000,000 loops (step=250000):
        - Cython: 18.61 s
        - Python: 31.30 s
    """
    signal = np.zeros((step, 1), dtype=DTYPE8, order='F')
    cdef int n = idx.size
    cdef DTYPE8_t[::1, :] signal_view = signal
    cdef DTYPE_t[:] idx_view = idx
    cdef int i
    with nogil:
        for i in range(n):
            signal_view[idx_view[i]][0] = 1
    return signal

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef check_zero_signal(DTYPE8_t[:] signal):
    """Check if the given signal contains only zeros.

    Parameters
    ----------
    signal : 1D array
        The signal to check.

    Returns
    -------
    bool
        True if the signal contains only zeros, False otherwise.

    Notes
    -----
    This method uses Cython to speed up the computation.

    Benchmark for 1,000,000 loops (1000/2000):
        - Cython: 1.45 s
        - Python: 3.67 s
    """
    cdef int n = signal.size
    cdef int i
    for i in range(n):
        if signal[i] == 1:
            return True
    return False
    



from pathlib import Path

import numpy as np
import pytest
from reacnetgenerator.tools import (
    read_species,
    calculate_rate,
)

p_cwd = Path(__file__).parent

def test_species():
    step_idx, n_species = read_species(p_cwd / 'methane.species')
    np.testing.assert_array_equal(step_idx, np.array([0, 1000, 2000, 3000, 4000, 5000, 6000, 7000], dtype=int))
    np.testing.assert_array_equal(n_species['[H]OO'], np.array([0, 0, 0, 1, 1, 1, 1, 1], dtype=int))

def test_rates():
    rates = calculate_rate(p_cwd / 'methane.species', p_cwd / 'methane.reactionabcd', np.eye(3) * 3.7601e1, 0.1)
    assert 'OO->O=O' in rates



# cython: language_level=3
# cython: linetrace=True
"""HMM Filter.

In order to filter noise, a two-state HMM was adopted, which can be
described as a transition matrix A, an emission matrix B, and an initial
state vector π. The existence of molecules can be converted into 0-1 signals.
In order to predict the state sequence according to the output sequence,
Viterbi Algorithm is used to acquire the path with the most likely hidden
state sequence V called Viterbi path.

References
----------
.. [1] Wang, L.-P.; McGibbon, R. T.; Pande, V. S.; Martínez, T. J. Automated
   discovery and refinement of reactive molecular dynamics pathways. J. Chem.
   Theory Comput. 2016, 12(2), 638-649.
.. [2] Rabiner, L. R. A trtorial on hidden Markov models and selected
   applications in speech recognition. Proc. IEEE 1989, 77(2), 257-286.
.. [3] Forney, G. D. The viterbi algorithm. Porc. IEEE 1973, 61(3), 268-278.
"""

import tempfile
from contextlib import ExitStack

import numpy as np
try:
    # hmmlearn v0.2.8 renamed MultinomialHMM to CategoricalHMM
    from hmmlearn.hmm import CategoricalHMM as MultinomialHMM
except ImportError:
    from hmmlearn.hmm import MultinomialHMM

from .utils import (
    WriteBuffer,
    appendIfNotNone,
    bytestolist,
    listtobytes,
    run_mp,
    SharedRNGData,
    read_compressed_block,
)
from .utils_np import idx_to_signal, check_zero_signal


class _HMMFilter(SharedRNGData):
    runHMM: bool
    getoriginfile: bool
    printfiltersignal: bool
    moleculetempfilename: str
    nproc: int
    temp1it: int
    p: np.ndarray
    a: np.ndarray
    b: np.ndarray
    step: int

    def __init__(self, rng):
        SharedRNGData.__init__(self, rng, ['runHMM', 'getoriginfile', 'printfiltersignal',
                                           'moleculetempfilename', 'nproc', 'temp1it', 'p', 'a', 'b', 'step'],
                               ['moleculetemp2filename', 'originfilename', 'hmmfilename', 'hmmit'])

    def filter(self):
        """HMM Filters.

        Timesteps of molecules are converted to a visible output sequence.
        O^m=(o_t^m) is given by o_t^m={1, if m exists; 0, otherwise}.
        Similarly, a hidden state sequence I^m=(i_t^m) is given by
        i_t^m={1, if m exists; 0, otherwise.}
        """
        if self.runHMM:
            self._initHMM()
        self._calhmm()
        self.returnkeys()

    def _initHMM(self):
        
        self._model = MultinomialHMM(n_components=2, algorithm="viterbi")
        self._model.startprob_ = self.p
        self._model.transmat_ = self.a
        self._model.emissionprob_ = self.b

    def _getoriginandhmm(self, item):
        line_c = item
        value = bytestolist(line_c[-1])
        origin = idx_to_signal(value, self.step)
        originbytes = listtobytes(
            origin) if self.getoriginfile else None
        hmmbytes = None
        if self.runHMM:
            hmmsignal = self._model.predict(origin).astype(np.int8)
            if check_zero_signal(hmmsignal) or self.printfiltersignal:
                hmmbytes = listtobytes(hmmsignal)
        return originbytes, hmmbytes, line_c

    def _calhmm(self):
        with WriteBuffer(tempfile.NamedTemporaryFile('wb', delete=False)) if self.getoriginfile or not self.runHMM else ExitStack() as fo, \
                WriteBuffer(tempfile.NamedTemporaryFile('wb', delete=False)) if self.runHMM else ExitStack() as fh, \
                open(self.moleculetempfilename, 'rb') as ft, \
                WriteBuffer(tempfile.NamedTemporaryFile('wb', delete=False)) as ft2:
            self.moleculetemp2filename = ft2.name
            if self.getoriginfile or not self.runHMM:
                assert not isinstance(fo, ExitStack)
                self.originfilename = fo.name
            else:
                self.originfilename = None
            if self.runHMM:
                assert not isinstance(fh, ExitStack)
                self.hmmfilename = fh.name
            else:
                self.hmmfilename = None
            results = run_mp(self.nproc, func=self._getoriginandhmm, l=read_compressed_block(ft),
                                nlines=4, total=self.temp1it, desc="HMM filter", unit="molecule")
            hmmit = 0
            for originbytes, hmmbytes, line_c in results:
                if originbytes is not None or hmmbytes is not None:
                    appendIfNotNone(fo, originbytes)
                    appendIfNotNone(fh, hmmbytes)
                    hmmit += 1
                    ft2.extend(line_c)
        self.hmmit = hmmit



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
#import os
#import sys
#sys.path.insert(0, os.path.abspath('..'))
from datetime import datetime


# -- Project information -----------------------------------------------------

project = 'ReacNetGenerator'
copyright = '2019-%d, East China Normal University' % datetime.now().year
author = 'Jinzhe Zeng'

# The full version, including alpha/beta/rc tags
#release = 'reacnetgenerator'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinxarg.ext',
    'myst_nb',
    'numpydoc',
    'sphinx-favicon',
    'deepmodeling_sphinx',
    "sphinxcontrib.bibtex",
]

# Add any paths that contain templates here, relative to this directory.
#templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'
html_logo = '_static/reacnetgen.svg'
html_static_path = ['_static']
html_js_files = ['https://unpkg.com/bilitube@0/dist/bilitube.min.js']
html_css_files = ['css/custom.css']
html_extra_path = ['report.html', 'fire.png', 'bundle.js', 'bundle.css']

html_theme_options = {
    "use_edit_page_button": True,
    "icon_links": [
        {"name": "Paper", "url": "https://doi.org/10.1039/C9CP05091D", "icon": "fa-solid fa-book-open", "type": "fontawesome"},
    ],
    "external_links": [
        {"name": "Report Loader", "url": "/report.html"}
    ],
    "github_url": "https://github.com/tongzhugroup/reacnetgenerator",
    "logo": {
        "text": "ReacNetGenerator",
        "alt_text": "ReacNetGenerator",
    },
}

html_context = {
    "github_user": "tongzhugroup",
    "github_repo": "reacnetgenerator",
    "github_version": "master",
    "doc_path": "docs",
}

myst_heading_anchors = 3

favicons = [
    {
        "rel": "icon",
        "static-file": "reacnetgen.svg",
        "type": "image/svg+xml",
    },
]

enable_deepmodeling = False

myst_enable_extensions = [
    "dollarmath",
    "colon_fence",
]
mathjax_path = "https://cdnjs.cloudflare.com/ajax/libs/mathjax/3.2.2/es5/tex-mml-chtml.min.js"
mathjax_options = {
    'integrity': 'sha512-6FaAxxHuKuzaGHWnV00ftWqP3luSBRSopnNAA2RvQH1fOfnF/A1wOfiUWF7cLIOFcfb1dEhXwo5VG3DAisocRw==',
    'crossorigin': 'anonymous',
}
mathjax3_config = {
    "loader": {"load": ['[tex]/mhchem']},
    "tex": {"packages": {'[+]': ['mhchem']}},
}

execution_mode = "off"
numpydoc_show_class_members = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

intersphinx_mapping = {
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "python": ("https://docs.python.org/", None),
    "ase": ("https://wiki.fysik.dtu.dk/ase/", None),
}

bibtex_bibfiles = ['refs.bib']

def run_apidoc(_):
    from sphinx.ext.apidoc import main
    import os
    import sys
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    module = os.path.join(cur_dir, "..", "reacnetgenerator")
    main(['-M', '--tocfile', 'api', '-H', 'Python API', '-o', os.path.join(cur_dir, "api"), module, '--force'])

def copy_report(app):
    import os
    from pathlib import Path

    import yaml
    from sphinx.util.fileutil import copy_asset_file
    from nodejs import node

    cur_dir = Path(__file__).parent.absolute()
    webpack = Path(cur_dir) / ".." / "reacnetgenerator" / "static" / "webpack"
    with open(webpack / ".yarnrc.yml") as f:
        yarn_path = Path(yaml.load(f, Loader=yaml.Loader)["yarnPath"])
    node.call([yarn_path], cwd=webpack)
    node.call([yarn_path, "start"], cwd=webpack, env={**os.environ, "REACNETGENERATOR_BUILDWEB": "1"})
    outdir = app.builder.outdir
    # first create the directory..
    os.makedirs(outdir, exist_ok=True)
    copy_asset_file(os.path.join(webpack, "bundle.html"), os.path.join(outdir, "report.html"))
    copy_asset_file(os.path.join(webpack, "bundle.js"), os.path.join(outdir, "bundle.js"))
    copy_asset_file(os.path.join(webpack, "bundle.css"), os.path.join(outdir, "bundle.css"))
    copy_asset_file(os.path.join(webpack, "fire.png"), os.path.join(outdir, "fire.png"))


def setup(app):
    app.connect('builder-inited', run_apidoc)
    app.connect('builder-inited', copy_report)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
# cython: language_level=3
# cython: linetrace=True
"""GUI version of ReacNetGenerator.
"""
import os
import tkinter as tk
import tkinter.filedialog as tkfd
import tkinter.messagebox as tkmb
import webbrowser
import base64
import pkg_resources
from multiprocessing import cpu_count

from . import ReacNetGenerator, __version__
from ._logging import logger


class GUI:
    """GUI class."""

    def __init__(self):
        """Init GUI class."""
        logger.info(__doc__)
        self._filename = ''

        self._top = tk.Tk()
        self._top.title(f"ReacNetGenerator {__version__}")
        self._top.columnconfigure(4, weight=2)

        self._filetype = tk.StringVar()
        self._filetype.set('lammpsbondfile')
        self._runhmm = tk.IntVar()
        self._runhmm.set(0)
        self._openpage = tk.IntVar()
        self._openpage.set(1)

        titleimage = tk.PhotoImage(
            data=base64.b64encode(
                pkg_resources.resource_string(
                    __name__, 'static/img-title.png')).decode())
        self._titlelb = tk.Label(self._top, image=titleimage)
        #self._titlelb.image = titleimage
        self._filenamelb = tk.Label(self._top, text="Trajectory File")
        self._filenameet = tk.Entry(self._top, width=45)
        self._openbtn = tk.Button(
            self._top, text="Select File", command=self._openfiles)
        self._filetypelb = tk.Label(self._top, text="File Type")
        self._filetyperbbond = tk.Radiobutton(
            text="LAMMPS Bond file", value="lammpsbondfile",
            variable=self._filetype)
        self._filetyperbtype = tk.Radiobutton(
            text="LAMMPS Dump file", value="lammpsdumpfile",
            variable=self._filetype)
        self._runhmmcb = tk.Checkbutton(
            text="HMM Filter", variable=self._runhmm, onvalue=1, offvalue=0)
        self._openpagecb = tk.Checkbutton(
            text="Show Report", variable=self._openpage, onvalue=1, offvalue=0)
        self._atomnamelb = tk.Label(self._top, text="Atomic Symbol")
        self._atomnameet = tk.Entry(self._top, width=18)
        self._atomnameet.insert(0, "C H O")
        self._runbtn = tk.Button(self._top, text="Analyze", command=self._run)
        self._nproclb = tk.Label(self._top, text="Processor Number")
        self._nprocet = tk.Entry(self._top, width=4)
        self._nprocet.insert(0, str(cpu_count()))

        self._titlelb.grid(row=0, column=0, columnspan=8, padx=20, pady=20)
        self._filenamelb.grid(row=1, column=0, columnspan=2, padx=5, pady=5)
        self._filenameet.grid(row=1, column=2, columnspan=5, padx=5, pady=5)
        self._openbtn.grid(row=1, column=7, columnspan=1, padx=5, pady=5)
        self._filetypelb.grid(row=2, column=0, columnspan=2, padx=5, pady=5)
        self._filetyperbbond.grid(
            row=2, column=2, columnspan=2, padx=5, pady=5)
        self._filetyperbtype.grid(
            row=2, column=4, columnspan=2, padx=5, pady=5)
        self._atomnamelb.grid(row=3, column=0, columnspan=2, padx=5, pady=5)
        self._atomnameet.grid(row=3, column=2, columnspan=2, padx=5, pady=5)
        self._runhmmcb.grid(row=2, column=6, columnspan=2, padx=5, pady=5)
        self._openpagecb.grid(row=3, column=6, columnspan=3, padx=5, pady=5)
        self._runbtn.grid(row=4, column=0, columnspan=8, padx=10, pady=10)
        self._nproclb.grid(row=3, column=4, padx=5, pady=5)
        self._nprocet.grid(row=3, column=5, padx=5, pady=5)

    def gui(self):
        """Start the GUI."""
        self._top.mainloop()

    @property
    def root(self):
        """Return the root of the GUI."""
        return self._top

    def _run(self):
        filename = self._filenameet.get()
        if os.path.exists(filename):
            try:
                reacnetgenerator = ReacNetGenerator(
                    inputfilename=self._filenameet.get(),
                    inputfiletype=self._filetype.get(),
                    atomname=self._atomnameet.get().split(),
                    runHMM=(self._runhmm.get() == 1),
                    nproc=int(self._nprocet.get()))
                reacnetgenerator.runanddraw()
                if self._openpage.get() == 1:
                    webbrowser.open_new(os.path.abspath(
                        reacnetgenerator.resultfilename))
            except Exception as e:
                logger.exception(e)
                tkmb.showerror("Error", str(e))
        else:
            logger.error("File not exsit.")
            tkmb.showerror("Error", "File not exsit.")

    def _openfiles(self):
        self._filename = tkfd.askopenfilename()
        if self._filename != '':
            self._filenameet.delete(0, tk.END)
            self._filenameet.insert(0, self._filename)


def gui():
    """Open GUI version of ReacNetGenerator."""
    GUI().gui()



import os
from pathlib import Path

import numpy as np
import pytest
from reacnetgenerator import ReacNetGenerator
from reacnetgenerator._detect import _Detect

p_inputs = Path(__file__).parent / 'inputs'


class TestDetect:
    """Test different detect format.
    
    All systems contain a single water molecule: H, H, O.
    """

    @pytest.fixture(autouse=True)
    def chdir(self, tmp_path):
        start_direcroty = os.getcwd()
        os.chdir(tmp_path)
        yield
        os.chdir(start_direcroty)

    @pytest.fixture(params=[
        # inputfiletype, inputfilename
        ("lammpsdumpfile", p_inputs / "water.dump"),
        ("dump", p_inputs / "water_pbc.dump"),
        ("lammpsbondfile", p_inputs / "water.bond"),
        ("xyz", p_inputs / "water.xyz"),
    ])
    def reacnetgen_param(self, request):
        return request.param

    @pytest.fixture()
    def reacnetgen(self, reacnetgen_param):
        rngclass = ReacNetGenerator(
            inputfiletype=reacnetgen_param[0],
            inputfilename=reacnetgen_param[1],
            atomname=["H", "O"],
            pbc=False,
        )
        yield rngclass

    def test_reacnetgen(self, reacnetgen):
        """Test main process of ReacNetGen."""
        _Detect.gettype(reacnetgen).detect()
        assert reacnetgen.N == 3
        np.testing.assert_array_equal(reacnetgen.atomtype, np.array([0, 0, 1], dtype=int))
        # assert this is a single molecule
        assert reacnetgen.temp1it == 1



# cython: language_level=3
# cython: linetrace=True
"""Obtain the version."""
from pkg_resources import DistributionNotFound, get_distribution
try:
    from ._version2 import version as __version__
except ImportError:
    try:
        __version__ = get_distribution('reacnetgenerator').version
    except DistributionNotFound:
        __version__ = 'Unknown'



# cython: language_level=3
# cython: linetrace=True
"""This module provides a logger for logging."""
import logging
import coloredlogs
from ._version import __version__

logger = logging.getLogger(__name__)
coloredlogs.install(
    fmt=f'%(asctime)s - ReacNetGenerator {__version__} - %(levelname)s: %(message)s',
    level=logging.INFO, milliseconds=True, logger=logger)

__all__ = ['logger']



"""Useful methods to futhur process ReacNetGenerator results."""
from collections import defaultdict
from pathlib import Path
from typing import Tuple, Dict, List, Union
from collections import Counter

import numpy as np
import ase


def read_species(specfile: Union[str, Path]) -> Tuple[np.ndarray, Dict[str, np.ndarray]]:
    """Read species from the species file (ends with .species).

    For accuracy, HMM filter should be disabled.

    Parameters
    ----------
    specfile : str or Path
        The species file.

    Returns
    -------
    step_idx : np.ndarray
        The index of the step.
    n_species : Dict[str, np.ndarray]
        The number of species in each step. The dict key is the species SMILES.

    Examples
    --------
    Plot the number of methane in each step.

    >>> from reacnetgenerator.tools import read_species
    >>> import matplotlib.pyplot as plt
    >>> step_idx, n_species = read_species('methane.species')
    >>> plt.plot(step_idx, n_species['[H]C([H])([H])[H]'])
    >>> plt.savefig("methane.svg")
    """
    step_idx = []
    n_species = defaultdict(lambda: defaultdict(int))
    with open(specfile) as f:
        ii = -1
        for ii, line in enumerate(f):
            s = line.split()
            step_idx.append(int(s[1].strip(':')))
            for ss, nn in zip(s[2::2], [int(x) for x in s[3::2]]):
                n_species[ss][ii] = nn
        else:
            nsteps = ii + 1
    n_species2 = {}
    for ss in n_species:
        n_species2[ss] = np.array([n_species[ss][ii] for ii in range(nsteps)], dtype=int)
    return np.array(step_idx, dtype=int), n_species2


def read_reactions(reacfile: Union[str, Path]) -> List[Tuple[int, Counter, str]]:
    """Read reactions from the reactions file (ends with .reaction or .reactionsabcd).

    For accuracy, HMM filter should be disabled.
    
    Parameters
    ----------
    reacfile : str or Path
        The reactions file.

    Returns
    -------
    occs : List[Tuple[int, Counter, str]]
        The number of occurences of each reaction. The tuple is (occurence, counter_reactants, reaction).
    """
    occs = []
    with open(reacfile) as f:
        for line in f:
            s = line.split()
            occs.append((int(s[0]), Counter(s[1].split('->')[0].split('+')), s[1]))
    return occs


def calculate_rate(specfile: Union[str, Path], reacfile: Union[str, Path], cell: np.ndarray, timestep: float) -> Dict[str, float]:
    """Calculate the rate constant of each reaction.

    The rate constants are calculated by the method developed in [1]_.
    The time interval of the trajectory is assumed to be uniform.

    Parameters
    ----------
    specfile : str
        The species file.
    reacfile : str
        The reactions file.
    cell : np.ndarray
        The cell with the shape (3, 3). Unit: Angstrom.
    timestep : float
        The time step. Unit: femtosecond.

    Returns
    -------
    rates : Dict[str, float]
        The rate of each reaction. The dict key is the reaction SMILES.
        The value is in unit of [(cm^3/mol)^(n-1)s^(-1)], where n is the reaction order.

    References
    ----------
    .. [1] Yanze Wu, Huai Sun, Liang Wu, Joshua D. Deetz, Extracting the mechanisms
       and kinetic models of complex reactions from atomistic simulation data, J.
       Comput. Chem. 40, 16, 1586-1592.

    Examples
    --------
    >>> cell = np.eye(3) * 3.7601e1 # in unit Angstrom
    >>> timestep = 0.1 # in unit fs
    >>> rates = calculate_rate('methane.species', 'methane.reactionabcd', cell, timestep)
    """
    ase_cell = ase.geometry.Cell(cell)
    
    timestep *= 10**-15 # fs to s
    #N, step_tot =read_species(specfile)
    step_idx, n_species = read_species(specfile)
    occs = read_reactions(reacfile)

    # time interval between two frames
    time_int = (step_idx[1] - step_idx[0]) * timestep
    # volume of the cell
    volume = ase_cell.volume
    volume *= 10**-24 # Ang^3 to cm^3
    volume_times_na = volume * ase.units.mol # V * NA
    
    rates = {}
    for occ, reacts, reactions in occs:
        # k = occ_tot / ( V * time_tot * c_tot )
        # c_tot = N_tot / (V * NA)
        n_react = np.array([n_species[kk] for kk in reacts.keys()])
        nu = np.array(list(reacts.values()))
        c_po = np.power(n_react / volume_times_na, np.repeat(nu, n_react.shape[1]).reshape(n_react.shape))
        c_tot = np.sum(np.prod(c_po, axis=0))
        k = occ / (volume_times_na * time_int * c_tot)
        rates[reactions] = k
    return rates



# cython: language_level=3
# cython: linetrace=True
"""Draw reaction network.

With the reaction matrix, a reaction network can be drawn. Here the NetworkX
package is used to make a graph which indicates reactions between species.[1]_
Fruchterman-Reingold force-directed algorithm is used to make layout of nodes
relate to reaction quantity and different colors and widths of lines are drawn
depending on reaction quantity.[2]_ The distance of two species in the network,
the color and thickness of the line between them are determined by the number
of their reactions, making the reaction network more intuitive.

References
----------
.. [1] Hagberg, A.; Swart, P.; Daniel, S. C. Exploring network structure,
   dynamics, and function using NetworkX; Los Alamos National Lab. (LANL), Los
   Alamos, NM (United States): 2008.
.. [2] Fruchterman, T. M.; Reingold, E. M. Graph drawing by force-directed
   placement. Software: Practice and experince. 1991, 21(11),1129-1164.
"""

import math
from io import StringIO
from itertools import permutations

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import scour.scour

from .utils import SCOUROPTIONS, SharedRNGData
from ._logging import logger


class _DrawNetwork(SharedRNGData):
    atomname: np.ndarray
    tablefilename: str
    imagefilename: str
    maxspecies: int
    species: list
    speciesfilter: list
    start_color: np.ndarray
    end_color: np.ndarray
    node_size: int
    node_color: str
    font_size: int
    widthcoefficient: float
    k: float
    pos: dict
    nolabel: bool
    showid: bool
    split: int

    def __init__(self, rng):
        SharedRNGData.__init__(self, rng, ['atomname', 'tablefilename', 'imagefilename', 'maxspecies', 'species', 'speciesfilter',
                                  'start_color', 'end_color', 'node_size', 'node_color', 'font_size', 'widthcoefficient', 'k', 'pos', 'nolabel',
                                  'showid', 'split'], [])

    def draw(self):
        """Draw the network."""
        self._draw()
        if self.split > 1:
            for st in range(self.split):
                self._draw(timeaxis=st)

    def _draw(self, timeaxis=None):
        table, name = self._readtable(
            self.tablefilename if timeaxis is None else f"{self.tablefilename}.{timeaxis}")
        species, showname = self._handlespecies(name)

        G = nx.DiGraph()
        idx = []
        for i, _ in enumerate(table):
            if name[i] in species:
                G.add_node(showname[name[i]])
                idx.append(i)
        for i, j in permutations(idx, 2):
            if table[i][j] > 0:
                G.add_weighted_edges_from(
                    [(showname[name[i]], showname[name[j]], table[i][j])])
        weights = np.array([math.log(G[u][v]['weight']+1)
                            for u, v in G.edges()])
        widths = weights / np.max(weights) * self.widthcoefficient * np.array(
            [0.5, 2])[(weights > np.max(weights) * 0.7)+0] if weights.size else np.zeros(0)
        colors = self.start_color + weights[:, np.newaxis] / np.max(weights) * (
            self.end_color - self.start_color) if weights.size else np.zeros(0)
        try:
            pos = nx.spring_layout(G,
                                   pos=self.pos if self.pos else None,
                                   fixed=list(self.pos) if self.pos else None,
                                   k=self.k)
            if pos:
                logger.info("The position of the species in the network is:")
                logger.info(pos)
            for with_labels in ([True] if not self.nolabel else [True, False]):
                nx.draw(
                    G, pos=pos, width=widths, node_size=self.node_size,
                    font_size=self.font_size, with_labels=with_labels,
                    edge_color=colors, node_color=[self.node_color]*len(pos))
                imagefilename = "".join(
                    (("" if with_labels else "nolabel_"), self.imagefilename))
                with StringIO() as stringio, open(imagefilename if timeaxis is None else f"{imagefilename}.{timeaxis}", 'w') as f:
                    plt.savefig(stringio, format='svg')
                    f.write(scour.scour.scourString(
                        stringio.getvalue(), SCOUROPTIONS))
                plt.close()
        except Exception as e:
            logger.exception(f"Error: cannot draw images. Details: {e}")

    def _readtable(self, filename):
        df = pd.read_csv(filename, sep=' ', index_col=0, header=0)
        return df.values, df.index

    def _handlespecies(self, name):
        species = self.species if self.species else name[:min(
            len(name), self.maxspecies)]
        # filter
        species = [spec for spec in species if spec not in self.speciesfilter]

        if self.showid:
            showname = dict([(v, u) for u, v in enumerate(species, start=1)])
        else:
            showname = dict([(u, u) for u in species])
        if species:
            logger.info("Species are:")
            for specname, n in showname.items():
                logger.info("{} {}".format(n, specname))
        return species, showname



import itertools

from tqdm import tqdm
import numpy as np

from .utils import (
    bytestolist,
    listtobytes,
    SharedRNGData,
    read_compressed_block,
)


class _mergeISO(SharedRNGData):
    miso: int
    moleculetempfilename: str
    temp1it: int

    def __init__(self, rng):
        SharedRNGData.__init__(
            self, rng, ['miso', 'moleculetempfilename'], ['temp1it'])

    def merge(self):
        if self.miso > 0:
            self._mergeISO()
        self.returnkeys()

    def _mergeISO(self):
        items = []
        with open(self.moleculetempfilename, 'rb') as ft:
            for item in itertools.zip_longest(*[read_compressed_block(ft)]*4):
                items.append(item)
        new_items = []
        _oldbitem = b''
        _oldbbond = b'0'
        _oldindex = []
        _oldfreq = 0
        for _bitem, _bbond0, _bbond1, _bindex in tqdm(sorted(items, key=lambda x: (x[0], x[1])), desc='Merge isomers:'):
            _index = bytestolist(_bindex)
            _freq = len(_index)
            if (_bitem == _oldbitem) and ((_bbond0 == _oldbbond[0]) or (self.miso > 1)):
                _oldindex = np.hstack((_oldindex, _index))
            else:
                if _oldbitem:
                    new_items.append(
                        [_oldbitem, *_oldbbond, listtobytes(_oldindex)])
                _oldbitem = _bitem
                _oldindex = _index
                if _freq > _oldfreq:
                    _oldbbond = (_bbond0, _bbond1)
        new_items.append([_oldbitem, *_oldbbond, listtobytes(_oldindex)])
        new_items.sort(key=lambda x: len(x[0]))
        self.temp1it = len(new_items)
        with open(self.moleculetempfilename, 'wb') as ft:
            for item in new_items:
                [ft.write(i) for i in item]



# cython: language_level=3
# cython: linetrace=True
"""Download trajectory before running."""

from typing import List

from .utils import SharedRNGData, download_multifiles

class DownloadData(SharedRNGData):
    urls: List[dict]

    def __init__(self, rng):
        SharedRNGData.__init__(self, rng, ["urls"], [])

    def download_files(self):
        download_multifiles(self.urls)



# cython: language_level=3
# cython: linetrace=True
"""Detect molecules.

There are two types of input files that could be imported by ReacNetGenerator,
the first part of necessary is the trajectory from reactive MD, the second
part can be the bond information normally given by simulation using ReaxFF.
In fact, atomic coordinates can be converted to the bond information with
the Open Babel software.[1]_ As a results, ReacNetGenerator can both processes ReaxFF
trajectories, AIMD trajectories, and other kinds of reactive trajectories.
With the bond information, molecules can be detected from atoms by Depth-first
search at every timestep.[2]_ By using this way, all molecules in a given
trajectory can be acquired. Molecules consisting of same atoms and bonds can
be considered as the same molecule.

References
----------
.. [1] O'Boyle, N. M.; Banck, M.; James, C. A.; Morley, C.; Vandermeersch, T.;
   Hutchison, G. Open Babel: An open chemical toolbox. J. Cheminf. 2011, 3(1),
   33-47.
.. [2] Tarjan, R. Depth-first search and linear graph algorithms. SIAM J. Comput.
   1972, 1 (2), 146-160.
"""

import tempfile
import fileinput
import operator
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from enum import Enum, auto
from typing import Optional, Tuple

import numpy as np
try:
    from openbabel import openbabel, __version__ as obversion
except ImportError:  # pragma: no cover
    raise ImportError("Open Babel 3.1.0 is required.")
from ase import Atom, Atoms
from packaging import version

from .dps import dps
from .utils import WriteBuffer, listtobytes, run_mp, SharedRNGData


if version.parse(obversion) < version.parse("3.1.0"):  # pragma: no cover
    raise ImportError("Open Babel 3.1.0 is required.")

openbabel.obErrorLog.StopLogging()


class _Detect(SharedRNGData, metaclass=ABCMeta):
    """Detect molecules.
    
    Parameters
    ----------
    rng: reacnetgenerator.ReacNetGenerator
        The ReacNetGenerator class.
    """
    subclasses = {}

    # type hints
    # TODO: we need a better way to communicate with parameters
    inputfilename: list
    atomname: np.ndarray
    stepinterval: int
    nproc: int
    pbc: bool
    cell: np.ndarray

    def __init__(self, rng):
        SharedRNGData.__init__(self, rng, ['inputfilename', 'atomname', 'stepinterval', 'nproc', 'pbc',
                                           'cell'],
                               ['N', 'atomtype', 'step', 'timestep', 'temp1it', 'moleculetempfilename'])

    @classmethod
    def gettype(cls, rng):
        """Get the class for the input file type.

        Now ReacNetGen support the following files:
            - lammpsbondfile: LAMMPS bond files, see http://lammps.sandia.gov/doc/fix_reax_bonds.html
            - lammpsdumpfile: LAMMPS dump files, see https://lammps.sandia.gov/doc/dump.html
            - xyz: xyz files

        Parameters
        ----------
        rng: reacnetgenerator.ReacNetGenerator
            The ReacNetGenerator class.
        
        Returns
        -------
        cls.subclasses[rng.inputfiletype](rng): class
            The _Detect class for specific file type.
        """
        if rng.inputfiletype not in cls.subclasses:
            raise ValueError(
                f"Unsupported input file type {rng.inputfiletype}")
        return cls.subclasses[rng.inputfiletype](rng)

    @classmethod
    def register_subclass(cls, message_type):
        """Register the file type, used as a decorator.
        
        Parameters
        ----------
        message_type: str
            The type name to register, such as `xyz`.
        
        Returns
        -------
        decorator: function
            The decorator that used for a subclass.

        Examples
        --------
        @_Detect.register_subclass("lammpsbondfile")
        class _DetectLAMMPSbond(_Detect):
        """
        def decorator(subclass):
            cls.subclasses[message_type] = subclass
            return subclass
        return decorator

    def detect(self):
        """Detect molecules."""
        self._readinputfile()
        self.returnkeys()

    def _readinputfile(self):
        """Read the input file."""
        d = defaultdict(list)
        timestep = {}
        with fileinput.input(files=self.inputfilename) as f:
            _steplinenum = self._readNfunc(f)
        with fileinput.input(files=self.inputfilename) as f:
            results = run_mp(self.nproc, func=self._readstepfunc, l=f, nlines=_steplinenum, return_num=True, interval=self.stepinterval,
                             desc="Read bond information and Detect molecules", unit="timestep")
            for molecules, (step, thetimestep) in results:
                for molecule in molecules:
                    d[molecule].append(step)
                timestep[step] = thetimestep
        self.temp1it = len(d)
        values_c = list(run_mp(self.nproc, func=self._compressvalue, l=d.values(
        ), unordered=False, desc="Save molecules", unit="molecule", total=self.temp1it))
        self._writemoleculetempfile((d.keys(), values_c))
        self.timestep = timestep
        self.step = len(timestep)

    def _compressvalue(self, x):
        return listtobytes(np.array(x))

    @abstractmethod
    def _readNfunc(self, f):
        pass

    @abstractmethod
    def _readstepfunc(self, item):
        pass

    def _connectmolecule(self, bond, level):
        # return list([b''.join((listtobytes(mol),
        #                        listtobytes(bondlist))) for mol, bondlist in zip(*dps(bond, level))])
        # int() because sometimes, the elements in bondlist are type numpy.int64 sometimes are int
        # which cause the different bytes results from same bond-network.
        mols, bondlists = dps(bond, level)
        return list([b''.join((listtobytes(mol),
                               listtobytes([(int(i[0]), int(i[1]))
                                           for i in bondlist]),
                               listtobytes([int(i[2]) for i in bondlist])
                               )) for mol, bondlist in zip(mols, bondlists)])

    def _writemoleculetempfile(self, d):
        with WriteBuffer(tempfile.NamedTemporaryFile('wb', delete=False)) as f:
            self.moleculetempfilename = f.name
            for mol in zip(*d):
                f.extend(mol)


@_Detect.register_subclass("bond")
@_Detect.register_subclass("lammpsbondfile")
class _DetectLAMMPSbond(_Detect):
    def _readNfunc(self, f):
        iscompleted = False
        N = None
        atomtype = None
        stepaindex = None
        index = -1
        for index, line in enumerate(f):
            if line[0] == '#':
                if line.startswith("# Number of particles"):
                    if iscompleted:
                        assert stepaindex is not None
                        steplinenum = index-stepaindex
                        break
                    else:
                        iscompleted = True
                        stepaindex = index
                    N = [int(s) for s in line.split() if s.isdigit()][0]
                    atomtype = np.zeros(N, dtype=int)
            else:
                s = line.split()
                assert atomtype is not None
                atomtype[int(s[0])-1] = int(s[1])-1
        else:
            steplinenum = index + 1
        assert N is not None and atomtype is not None
        self.N = N
        self.atomtype = atomtype
        return steplinenum

    def _readstepfunc(self, item):
        step, lines = item
        bond: list[Optional[Tuple[int]]] = [None]*self.N
        level: list[Optional[Tuple[int]]] = [None]*self.N
        timestep = None
        for line in lines:
            if line:
                if line[0] == "#":
                    if line.startswith("# Timestep"):
                        timestep = int(line.split()[-1])
                else:
                    s = line.split()
                    s0 = int(s[0])-1
                    s2 = int(s[2])
                    bond[s0] = tuple(map(self._get_idx, s[3:3+s2]))
                    level[s0] = tuple(map(self._get_bo, s[4+s2:4+2*s2]))
        molecules = self._connectmolecule(bond, level)
        assert timestep is not None
        return molecules, (step, timestep)

    @staticmethod
    def _get_idx(x):
        return int(x) - 1

    @staticmethod
    def _get_bo(x):
        return max(1, round(float(x)))


class _DetectCrd(_Detect):
    def _getbondfromcrd(self, step_atoms, cell):
        atomnumber = len(step_atoms)
        # Use openbabel to connect atoms
        mol = openbabel.OBMol()
        mol.BeginModify()
        for idx, (num, position) in enumerate(zip(step_atoms.get_atomic_numbers(), step_atoms.positions)):
            a = mol.NewAtom(idx)
            a.SetAtomicNum(int(num))
            a.SetVector(*position)
        # Apply period boundry conditions
        # openbabel#1853, supported in v3.1.0
        if self.pbc:
            uc = openbabel.OBUnitCell()
            uc.SetData(
                openbabel.vector3(cell[0][0], cell[0][1], cell[0][2]),
                openbabel.vector3(cell[1][0], cell[1][1], cell[1][2]),
                openbabel.vector3(cell[2][0], cell[2][1], cell[2][2]),
            )
            mol.CloneData(uc)
            mol.SetPeriodicMol()
        mol.ConnectTheDots()
        mol.PerceiveBondOrders()
        mol.EndModify()
        bond = [[] for i in range(atomnumber)]
        bondlevel = [[] for i in range(atomnumber)]
        for b in openbabel.OBMolBondIter(mol):
            s1 = b.GetBeginAtom().GetId()
            s2 = b.GetEndAtom().GetId()
            level = b.GetBondOrder()
            if level == 5:
                # aromatic, 5 in openbabel but 12 in rdkit
                level = 12
            bond[s1].append(s2)
            bond[s2].append(s1)
            bondlevel[s1].append(level)
            bondlevel[s2].append(level)
        return bond, bondlevel


@_Detect.register_subclass("dump")
@_Detect.register_subclass("lammpsdumpfile")
class _DetectLAMMPSdump(_DetectCrd):
    class LineType(Enum):
        """Line type in the LAMMPS dump files."""

        TIMESTEP = auto()
        ATOMS = auto()
        NUMBER = auto()
        BOX = auto()
        OTHER = auto()

        @classmethod
        def linecontent(cls, line):
            """Return line content."""
            if line.startswith("ITEM: TIMESTEP"):
                return cls.TIMESTEP
            if line.startswith("ITEM: ATOMS"):
                return cls.ATOMS
            if line.startswith("ITEM: NUMBER OF ATOMS"):
                return cls.NUMBER
            if line.startswith("ITEM: BOX"):
                return cls.BOX
            return cls.OTHER

    def _readNfunc(self, f):
        iscompleted = False
        N = None
        linecontent = None
        atomtype = None
        index = -1
        stepaindex = None
        for index, line in enumerate(f):
            if line.startswith("ITEM:"):
                linecontent = self.LineType.linecontent(line)
                if linecontent == self.LineType.ATOMS:
                    keys = line.split()
                    self.id_idx = keys.index('id') - 2
                    self.tidx = keys.index('type') - 2
                    self.xidx = keys.index('x') - 2
                    self.yidx = keys.index('y') - 2
                    self.zidx = keys.index('z') - 2
            else:
                assert linecontent is not None
                if linecontent == self.LineType.NUMBER:
                    if iscompleted:
                        assert stepaindex is not None
                        steplinenum = index-stepaindex
                        break
                    else:
                        iscompleted = True
                        stepaindex = index
                    N = int(line.split()[0])
                    atomtype = np.zeros(N, dtype=int)
                elif linecontent == self.LineType.ATOMS:
                    s = line.split()
                    assert atomtype is not None
                    atomtype[int(s[self.id_idx])-1] = int(s[self.tidx])-1
        else:
            steplinenum = index + 1
        assert N is not None and atomtype is not None
        self.N = N
        self.atomtype = atomtype
        return steplinenum

    def _readstepfunc(self, item):
        step, lines = item
        step_atoms = []
        ss = []
        linecontent = None
        timestep = None
        for line in lines:
            if line:
                if line.startswith("ITEM:"):
                    linecontent = self.LineType.linecontent(line)
                else:
                    if linecontent is None:
                        raise ValueError("LAMMPS dump file format error")
                    elif linecontent == self.LineType.ATOMS:
                        s = line.split()
                        step_atoms.append(
                            (int(s[self.id_idx]),
                             Atom(
                                 self.atomname[int(s[self.tidx]) - 1],
                                 (float(s[self.xidx]), float(s[self.yidx]), float(s[self.zidx])))))
                    elif linecontent == self.LineType.TIMESTEP:
                        timestep = step, int(line.split()[0])
                    elif linecontent == self.LineType.BOX:
                        s = line.split()
                        ss.append(list(map(float, s)))
        assert timestep is not None
        ss = np.array(ss)
        if ss.shape[1] > 2:
            xy = ss[0][2]
            xz = ss[1][2]
            yz = ss[2][2]
        else:
            xy, xz, yz = 0., 0., 0.
        xlo = ss[0][0] - min(0., xy, xz, xy+xz)
        xhi = ss[0][1] - max(0., xy, xz, xy+xz)
        ylo = ss[1][0] - min(0., yz)
        yhi = ss[1][1] - max(0., yz)
        zlo = ss[2][0]
        zhi = ss[2][1]
        boxsize = np.array([[xhi-xlo, 0., 0.],
                            [xy, yhi-ylo, 0.],
                            [xz, yz, zhi-zlo]])
        _, step_atoms = zip(*sorted(step_atoms, key=operator.itemgetter(0)))
        step_atoms = Atoms(step_atoms)
        bond, level = self._getbondfromcrd(step_atoms, boxsize)
        molecules = self._connectmolecule(bond, level)
        return molecules, timestep


@_Detect.register_subclass("xyz")
class _Detectxyz(_DetectCrd):
    """xyz file. Two frames are connected. Cell information must be inputed by user."""

    def _readNfunc(self, f):
        atomname_dict = dict(
            zip(self.atomname.tolist(), range(self.atomname.size)))
        N = None
        atomtype = None
        for index, line in enumerate(f):
            s = line.split()
            if index == 0:
                N = int(line.strip())
                atomtype = np.zeros(N, dtype=int)
            elif N is None:
                assert N is not None  # pragma: no cover
            elif index > N+1:
                break
            elif index > 1:
                assert atomtype is not None
                atomtype[index-2] = atomname_dict[s[0]]
        assert N is not None and atomtype is not None
        self.N = N
        self.atomtype = atomtype
        steplinenum = N + 2
        return steplinenum

    def _readstepfunc(self, item):
        step, lines = item
        timestep = step, step
        step_atoms = []
        boxsize = self.cell
        if self.pbc and boxsize is None:
            raise RuntimeError("No cell information is given.")
        for index, line in enumerate(lines):
            s = line.split()
            if index > 1:
                step_atoms.append(
                    (index-1, Atom(s[0], tuple((float(x) for x in s[1:4])))))
        _, step_atoms = zip(*sorted(step_atoms, key=operator.itemgetter(0)))
        step_atoms = Atoms(step_atoms)
        bond, level = self._getbondfromcrd(step_atoms, boxsize)
        molecules = self._connectmolecule(bond, level)
        return molecules, timestep



# cython: language_level=3
# cython: linetrace=True
import argparse
import textwrap
from typing import List
from . import __version__
from ._detect import _Detect 

def main_parser() -> argparse.ArgumentParser:
    """Returns main parser.
    
    Returns
    -------
    argparse.ArgumentParser
        reacnetgenerator cli parser
    """
    parser = argparse.ArgumentParser(
        description='ReacNetGenerator is an automatic reaction network generator for reactive molecular dynamics simulation.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog=textwrap.dedent("""\
        Examples:
            reacnetgenerator --type bond -i bonds.reaxc -a C H O --nohmm
        """))
    parser.add_argument(
        '-i', '--inputfilename', nargs='*',
        help='Input trajectory file(s), e.g. bonds.reaxc; If multiple file are given, they are concatenated.',
        required=True)
    parser.add_argument(
        '-a', '--atomname',
        help='Atomic names in the trajectory, e.g. C H O',
        nargs='*', required=True)
    parser.add_argument(
        '--nohmm',
        help=('Process trajectory without Hidden Markov Model (HMM). If one wants to enable HMM, firstly '
              'read the related section in the paper'),
        action="store_true")
    parser.add_argument(
        '--miso', help='Merge the isomers',type=int, default=0)
    parser.add_argument(
        '--dump',
        help='(deprecated) Process the LAMMPS dump file. Please use `--type dump` instead.',
        action="store_true")
    parser.add_argument(
        '--type', '-t',
        help='Input file type',
        choices=list(_Detect.subclasses.keys()),
        default='lammpsbondfile')
    parser.add_argument(
        '--nopbc',
        help='Disable PBC.',
        action="store_true")
    parser.add_argument(
        '--cell', '-c', nargs='+', type=float,
        help=('Cell information if PBC is enabled and the input file does not contain cell information. '
              'The cell information should be in 3x3 matrix format.'))
    parser.add_argument(
        '-n', '-np', '--nproc', help='Number of processes, by default using all processes', type=int)
    parser.add_argument(
        '-s', '--selectatoms',
        help='Select element(s) in the reaction network, e.g. C', nargs='*')
    parser.add_argument(
        '--stepinterval',
        help='Step interval, i.e. read every N step from the trajectory',
        type=int, default=1)
    parser.add_argument(
        '--split',
        help='Split number for the time axis; the whole trajectroy will be splited into N parts for analysis',
        type=int, default=1)
    parser.add_argument(
        '--maxspecies', help='Max number of nodes (species) in the network', type=int, default=20)
    parser.add_argument(
        '--matrixa',
        help='Transition matrix A of HMM parameters',
        type=float, nargs=4, default=[0.999, 0.001, 0.001, 0.999])
    parser.add_argument(
        '--matrixb',
        help='Emission matrix B of HMM parameters',
        type=float, nargs=4, default=[0.6, 0.4, 0.4, 0.6])
    parser.add_argument(
        '--urls', action='append', nargs=2, type=str,
        help='Download files before analysis, in the format of `--urls filename url`')
    parser.add_argument('--version', action='version', version='ReacNetGenerator v%s' % __version__)
    return parser


def _commandline():
    parser = main_parser()
    args = parser.parse_args()
    from .reacnetgen import ReacNetGenerator
    import numpy as np
    ReacNetGenerator(
        inputfilename=args.inputfilename, atomname=args.atomname,
        miso=args.miso,
        runHMM=not args.nohmm,
        inputfiletype=('lammpsdumpfile' if args.dump else args.type),
        nproc=args.nproc, selectatoms=args.selectatoms,
        stepinterval=args.stepinterval,
        split=args.split,
        maxspecies=args.maxspecies,
        urls=[{"fn": url[0], "url": url[1]}
              for url in args.urls] if args.urls else None,
        a=np.array(args.matrixa).reshape((2, 2)),
        b=np.array(args.matrixb).reshape((2, 2)),
        pbc=not args.nopbc,
        cell=args.cell,
    ).runanddraw()


def parm2cmd(pp: dict) -> List[str]:
    """Converts a parameter dictionary to command line arguments.
    
    Parameters
    ----------
    pp : dict
        Parameter dictionary

    Returns
    -------
    List[str]
        Command line arguments
    """
    commands = ['reacnetgenerator', '-i',
                pp['inputfilename'], '-a', *pp['atomname']]
    if not pp.get('runHMM', True):
        commands.append('--nohmm')
    if pp['inputfiletype']:
        commands.extend(('--type', pp['inputfiletype']))
    if pp['atomname']:
        commands.extend(('-s', pp['atomname'][0]))
    if pp.get('urls', []):
        commands.extend(
            ('--urls', pp['urls'][0]['fn'], pp['urls'][0]['url'][0]))
    if pp.get('a', []):
        commands.extend(
            ('--matrixa', str(pp['a'][0][0]), str(pp['a'][0][1]), str(pp['a'][1][0]), str(pp['a'][1][1])))
    if pp.get('b', []):
        commands.extend(
            ('--matrixb', str(pp['b'][0][0]), str(pp['b'][0][1]), str(pp['b'][1][0]), str(pp['b'][1][1])))
    if pp.get('cell', []):
        commands.extend(
            ('--cell', str(pp['cell'][0]), str(pp['cell'][1]), str(pp['cell'][2])))
    for ii in ['nproc', 'selectatoms', 'stepinterval', 'split', 'maxspecies']:
        if pp.get(ii, None):
            commands.extend(("--{}".format(ii), str(pp[ii])))
    return commands



# cython: language_level=3
# cython: linetrace=True
"""Collect paths.

To produce a reaction network, every molecule (species) should be treated as a
node in the network. Therefore, all detected species are indexed by canonical
SMILES to guarantee its uniqueness. Isomers are also identified according to
SMILES codes._[1] The VF2 algorithm can be also used to identify isomers, which is
an option in ReacNetGenerator._[2] After filtering out noise, the reaction path of atoms
and the number of intermolecular reactions can be calculated.

References
----------
.. [1] Landrum, G. RDKit: Open-Source Cheminformatics Software 2016.
.. [2] Cordella, L. P.; Foggia, P.; Sansone, C.; Vento, M. A (Sub)Graph
   Isomorphism Algorith for Matching Large Graphs. IEEE Trans. Pattern Analysis
   and Machine Intelligence 2004, 26, 1367-1372.
"""

import itertools
from abc import ABCMeta, abstractmethod
from collections import Counter, defaultdict

import networkx as nx
import networkx.algorithms.isomorphism as iso
import numpy as np
from rdkit import Chem
from tqdm import tqdm

from ._reaction import ReactionsFinder
from .utils import (
    WriteBuffer,
    bytestolist,
    listtostirng,
    run_mp,
    SharedRNGData,
    read_compressed_block,
)


class _CollectPaths(SharedRNGData, metaclass=ABCMeta):
    runHMM: bool
    N: int
    step: int
    atomname: np.ndarray
    originfilename: str
    hmmfilename: str
    moleculefilename: str
    moleculetemp2filename: str
    atomroutefilename: str
    nproc: int
    hmmit: int
    atomtype: np.ndarray
    selectatoms: list
    split: int
    mname: np.ndarray

    def __init__(self, rng):
        SharedRNGData.__init__(self, rng, ["runHMM", "N", "step", "atomname", "originfilename", "hmmfilename", "moleculefilename",
                                           "moleculetemp2filename", "atomroutefilename", "nproc", "hmmit", "atomtype",
                                           "selectatoms", "split"], ["mname", "atomnames", "allmoleculeroute", "splitmoleculeroute"])

    @staticmethod
    def getstype(rng):
        """Get a class for different methods.

        Following methonds are used to identify isomers:
        * SMILES (default)
        * VF2
        """
        if rng.SMILES:
            return _CollectSMILESPaths(rng)
        return _CollectMolPaths(rng)

    def collect(self):
        """Collect paths."""
        self.atomnames = self.atomname[self.atomtype]
        self._printmoleculename()
        atomeach, conflict = self._getatomeach()
        self.allmoleculeroute = self._printatomroute(atomeach)
        if self.split > 1:
            splittime = np.array_split(np.arange(self.step), self.split)
            self.splitmoleculeroute = list([self._printatomroute(
                atomeach[:, st], timeaxis=i) for i, st in enumerate(splittime)])
        self.returnkeys()
        ReactionsFinder(self.rng).findreactions(atomeach.T, conflict.T)

    @abstractmethod
    def _printmoleculename(self):
        pass

    def _getatomeach(self):
        """Values in atomeach starts from 1."""
        atomeach = np.zeros((self.N, self.step), dtype=int)
        conflict = np.zeros((self.N, self.step), dtype=int)
        with open(self.hmmfilename if self.runHMM else self.originfilename, 'rb') as fh, \
             open(self.moleculetemp2filename, 'rb') as ft:
            for i, (linehz, linetz) in enumerate(tqdm(zip(read_compressed_block(fh), itertools.zip_longest(*[read_compressed_block(ft)] * 4)),
                                                      total=self.hmmit, desc="Analyze atoms", unit="molecule"), start=1):
                lineh = bytestolist(linehz)
                atom = np.array(bytestolist(linetz[0]))
                index = np.where(lineh)[0]
                if index.size:
                    conflict[np.nonzero(atomeach[atom[:, None], index])] = 1
                    atomeach[atom[:, None], index] = i
        return atomeach, conflict

    def _getatomroute(self, item):
        i, (atomeachi, atomtypei) = item
        atomeachi = atomeachi[np.nonzero(atomeachi)[0]]
        if atomeachi.size:
            time = np.concatenate([
                np.zeros((1,), dtype=int),
                np.nonzero(np.diff(atomeachi))[0]+1,
                ])
            route = atomeachi[time]
        else:
            time = np.zeros(0, dtype=int)
            route = np.zeros(0, dtype=int)
        moleculeroute = np.dstack((route[:-1], route[1:]))[
            0] if self.atomname[atomtypei] in self.selectatoms else np.zeros((0, 2), dtype=int)
        names = self.mname[route-1]
        # Atom {idx}: {time} {SMILES} -> {time} {SMILES} -> ...
        routestr = f"Atom {i} {self.atomname[atomtypei]}: " + \
            " -> ".join([f"{tt} {name}" for tt, name in zip(time, names)])
        return moleculeroute, routestr

    def _printatomroute(self, atomeach, timeaxis=None):
        """For analysis without HMM, we may not need to use np.unique"""
        with WriteBuffer(open(self.atomroutefilename if timeaxis is None else f"{self.atomroutefilename}.{timeaxis}", 'w'), sep='\n') as f:
            allmoleculeroute = []
            if not self.runHMM:
                have_added = {}
            else:
                have_added = None
            results = run_mp(self.nproc, func=self._getatomroute, l=zip(atomeach, self.atomtype), return_num=True, start=1, unordered=False,
                                total=self.N, desc="Collect reaction paths" if timeaxis is None else f"Collect reaction paths {timeaxis}", unit="atom")
            for ii, (moleculeroute, routestr) in enumerate(results):
                f.append(routestr)
                if moleculeroute.size > 0:
                    if not self.runHMM:
                        # check whether repeated or not if analyzing without HMM
                        for rr in moleculeroute:
                            tpr = tuple(rr)
                            assert have_added is not None
                            if have_added.get(tpr, atomeach.shape[0]) >= ii:
                                have_added[tpr] = ii
                                allmoleculeroute.append(rr.reshape(1,2))
                    else:
                        allmoleculeroute.append(moleculeroute)
        allmoleculeroute = np.concatenate(allmoleculeroute) if allmoleculeroute else np.zeros((0, 2), dtype=int)
        if self.runHMM and allmoleculeroute.size:
            allmoleculeroute = np.unique(allmoleculeroute, axis=0)
        return allmoleculeroute

    def convertSMILES(self, atoms, bonds):
        """Convert atoms and bonds information to SMILES.
        
        Raises
        ------
        ValueError
            (RDKit error) Maximum BFS search size exceeded.
        """
        m = Chem.RWMol(Chem.MolFromSmiles(''))
        d = {}
        for name, number in zip(self.atomnames[atoms], atoms):
            d[number] = m.AddAtom(Chem.Atom(name))
        for atom1, atom2, level in bonds:
            m.AddBond(d[atom1], d[atom2], Chem.BondType(level))
        name = Chem.MolToSmiles(m)
        return name

    def _getatomsandbonds(self, line):
        atoms = np.array(bytestolist(line[0]), dtype=int)
        pairs = bytestolist(line[1])
        levels = bytestolist(line[2])
        bonds = [[*pair, level] for pair, level in zip(pairs, levels)]
        return atoms, bonds


class _CollectMolPaths(_CollectPaths):
    """VF2 is used to identify isomers.
    
    If SMILES is failed to generate, fallback to the name like CxHyOz.
    """
    def _printmoleculename(self):
        mname = []
        d = defaultdict(list)
        em = iso.numerical_edge_match(['atom', 'level'], ["None", 1])
        # idx for unknown SMILES
        self.n_unknown = 0
        with WriteBuffer(open(self.moleculefilename, 'w'), sep='\n') as fm, \
             open(self.moleculetemp2filename, 'rb') as ft:
            for line in tqdm(itertools.zip_longest(*[read_compressed_block(ft)] * 4),
                            total=self.hmmit, desc="Indentify isomers", unit="molecule"):
                atoms, bonds = self._getatomsandbonds(line)
                molecule = Molecule(self, atoms, bonds)
                for isomer in d[str(molecule)]:
                    if isomer.isomorphic(molecule, em):
                        molecule.smiles = isomer.smiles
                        break
                else:
                    d[str(molecule)].append(molecule)
                mname.append(molecule.smiles)
                fm.append(listtostirng(
                    (molecule.smiles, atoms, bonds), sep=(' ', ';', ',')))
        self.mname = np.array(mname)


class _CollectSMILESPaths(_CollectPaths):
    def _printmoleculename(self):
        mname = []
        d = defaultdict(list)
        em = iso.numerical_edge_match(['atom', 'level'], ["None", 1])
        self.n_unknown = 0
        with WriteBuffer(open(self.moleculefilename, 'w'), sep='\n') as fm, \
             open(self.moleculetemp2filename, 'rb') as ft:
            results = run_mp(self.nproc, func=self._calmoleculeSMILESname, l=read_compressed_block(ft), unordered=False,
                                nlines=4, total=self.hmmit, desc="Indentify isomers", unit="molecule")
            for name, atoms, bonds in results:
                if name is None:
                    # SMILES failed, fallback to VF2 identify isomers
                    molecule = Molecule(self, atoms, bonds)
                    # directly raise ValueError to save time
                    def _raise_anyway(*args, **kwargs):
                        raise ValueError("Maximum BFS search size exceeded.")
                    molecule._convertSMILES = _raise_anyway
                    for isomer in d[str(molecule)]:
                        if isomer.isomorphic(molecule, em):
                            molecule.smiles = isomer.smiles
                            break
                    else:
                        d[str(molecule)].append(molecule)
                    name = molecule.smiles
                mname.append(name)
                fm.append(listtostirng(
                    (name, atoms, bonds), sep=(' ', ';', ',')))
        self.mname = np.array(mname)

    def _calmoleculeSMILESname(self, item):
        line = item
        atoms, bonds = self._getatomsandbonds(line)
        try:
            name = self.convertSMILES(atoms, bonds)
        except ValueError:
            # fallback to VF2
            name = None
        return name, atoms, bonds


class Molecule:
    """A molecule class for isomer identification."""
    def __init__(self, cmp, atoms, bonds):
        self.cmp = cmp
        self.atoms = atoms
        self.bonds = bonds
        self._atomtypes = cmp.atomtype[atoms]
        self._atomnames = cmp.atomnames[atoms]
        self.graph = self._makemoleculegraph()
        counter = Counter(self._atomnames)
        self.name = "".join(map(lambda atomname: f"{atomname}{counter[atomname]}",
                                cmp.atomname))
        self._smiles = None
        self._convertSMILES = cmp.convertSMILES

    def __str__(self):
        return self.name

    @property
    def smiles(self):
        """Return SMILES of a molecule."""
        if self._smiles is None:
            try:
                self._smiles = self._convertSMILES(self.atoms, self.bonds)
            except ValueError:
                # when RDKit error: Maximum BFS search size exceeded
                # fallback to the name of the molecule
                # blank should be avoided
                self._smiles = self.name + f"_unknownSMILES_{self.cmp.n_unknown}"
                self.cmp.n_unknown += 1
        return self._smiles

    @smiles.setter
    def smiles(self, value):
        self._smiles = value

    def _makemoleculegraph(self):
        graph = nx.Graph()
        for line in self.bonds:
            graph.add_edge(line[0], line[1], level=line[2])
        for atomnumber, atomtype in zip(self.atoms, self._atomtypes):
            graph.add_node(atomnumber, atom=atomtype)
        return graph

    def isomorphic(self, mol, em):
        """Return whether two molecules are isomorphic."""
        return nx.is_isomorphic(self.graph, mol.graph, em)



# cython: language_level=3
# cython: linetrace=True
"""Provide utils for ReacNetGenerator."""


from contextlib import ExitStack
import os
import shutil
import itertools
import pickle
import hashlib
import asyncio
from multiprocessing import Pool, Semaphore
from typing import (
    Any,
    BinaryIO,
    Callable,
    Iterable,
    List,
    Optional,
    Tuple,
    IO,
    AnyStr,
    Generator,
    Union,
    TYPE_CHECKING,
)

import lz4.frame
import numpy as np
import requests
from requests.adapters import HTTPAdapter
from tqdm import tqdm

from ._logging import logger

if TYPE_CHECKING:
    import multiprocessing.pool
    import multiprocessing.synchronize
    import reacnetgenerator


class WriteBuffer:
    """Store a buffer for writing files.

    It is expensive to write to a file, so we need to make a buffer.
    
    Parameters
    ----------
    f: fileObject
        The file object to write.
    linenumber: int, default: 1200
        The number of contents to store in the buffer. The buffer will be flushed
        if it exceeds the set number.
    sep: str or bytes, default: None
        The separator for contents. If None (default), there will be no separator.
    """
    def __init__(self, f: IO, linenumber: int = 1200, sep: Optional[AnyStr] = None) -> None:
        self.f = f
        if sep is not None:
            self.sep = sep
        elif f.mode == 'w':
            self.sep = ''
        elif f.mode == 'wb':
            self.sep = b''
        else:
            raise RuntimeError("File mode should be w or wb!")
        self.linenumber = linenumber
        self.buff = []
        self.name = self.f.name

    def append(self, text: AnyStr) -> None:
        """Append a text.
        
        Parameters
        ----------
        text: str or bytes
            The text to be appended.
        """
        self.buff.append(text)
        self.check()

    def extend(self, text: Iterable[AnyStr]) -> None:
        """Extend texts.

        Parameters
        ----------
        text: list of strs or bytes
            Texts to be extended.
        """
        self.buff.extend(text)
        self.check()

    def check(self) -> None:
        """Check if the number of stored contents exceeds.
        
        If so, the buffer will be flushed.
        """
        if len(self.buff) > self.linenumber:
            self.flush()

    def flush(self) -> None:
        """Flush the buffer."""
        if self.buff:
            self.f.writelines([self.sep.join(self.buff), self.sep])
            self.buff[:] = []

    def __enter__(self) -> "WriteBuffer":
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.flush()
        self.f.__exit__(exc_type, exc_value, traceback)


def appendIfNotNone(f: Union[WriteBuffer, ExitStack], wbytes: Optional[AnyStr]) -> None:
    """Append a line to a file if the line is not None.
    
    Parameters
    ----------
    f: WriteBuffer
        The file to write.
    wbytes: str or bytes
        The line to write.
    """
    if wbytes is not None:
        assert not isinstance(f, ExitStack)
        f.append(wbytes)


def produce(semaphore: "multiprocessing.synchronize.Semaphore",
            plist: Iterable[Any],
            parameter: Any) -> Generator[Tuple[Any, Any], None, None]:
    """Item producer with a semaphore.
    
    Prevent large memory usage due to slow IO.
    
    Parameters
    ----------
    semaphore: multiprocessing.Semaphore
        The semaphore to acquire.
    plist: list of objects
        The list of items to be passed.
    parameter: object
        The parameter yielded with each item.

    Yields
    ------
    item: object
        The item to be yielded.
    parameter: object
        The parameter yielded with each item.
    """
    for item in plist:
        semaphore.acquire()
        if parameter is not None:
            item = (item, parameter)
        yield item


def compress(x: Union[str, bytes]) -> bytes:
    """Compress the line.

    This function reduces IO overhead to speed up the program. The functions will
    use lz4 to compress, since lz4 has better performance
    that any others.

    The compressed format is size + data + size + data + ..., where size is a 64-bit
    little-endian integer.

    Parameters
    ----------
    x: str or bytes
        The line to compress.

    Returns
    -------
    bytes
        The compressed line, with a linebreak in the end.
    """
    if isinstance(x, str):
        x = x.encode()
    compress_block = lz4.frame.compress(x, compression_level=0)
    length_bytes = len(compress_block).to_bytes(64, byteorder='little')
    return length_bytes + compress_block


def decompress(x: bytes, isbytes: bool = False) -> Union[str, bytes]:
    """Decompress the line.
    
    Parameters
    ----------
    x: bytes
        The line to decompress.
    isbytes: bool, optional, default: False
        If the decompressed content is bytes. If not, the line will be decoded.
    
    Returns
    -------
    str or bytes
        The decompressed line.
    """
    x = lz4.frame.decompress(x[64:])
    if isbytes:
        return x
    return x.decode()


def listtobytes(x: Any) -> bytes:
    """Convert an object to a compressed line.
    
    Parameters
    ----------
    x: object
        The object to convert, such as numpy.ndarray.
    
    Returns
    -------
    bytes
        The compressed line.
    """
    return compress(pickle.dumps(x))


def read_compressed_block(f: BinaryIO) -> Generator[bytes, None, None]:
    """Read compressed binary file, assuming the format is size + data + size + data + ...
    
    Parameters
    ----------
    f: fileObject
        The file object to read.

    Yields
    ------
    data: bytes
        The compressed block.
    """
    while True:
        sizeb = f.read(64)
        if not sizeb:
            break
        size = int.from_bytes(sizeb, byteorder='little')
        yield sizeb + f.read(size)


def bytestolist(x: bytes) -> Any:
    """Convert a compressed line to an object.
    
    Parameters
    ----------
    x: bytes
        The compressed line.
    
    Returns
    -------
    object
        The decompressed object.
    """
    data = decompress(x, isbytes=True)
    assert isinstance(data, bytes)
    return pickle.loads(data)


def listtostirng(l: Union[str, list, tuple, np.ndarray], sep: Union[List[str], Tuple[str, ...]]) -> str:
    """Convert a list to string, that is easier to store.

    Parameters
    ----------
    l: str or array-like
        The list to convert, which can contain any number of dimensions.
    sep: list of strs
        The seperators for each dimension.

    Returns
    -------
    str
        The converted string.
    """
    if isinstance(l, str):
        return l
    if isinstance(l, (list, tuple, np.ndarray)):
        return sep[0].join(map(lambda x: listtostirng(x, sep[1:]), l))
    return str(l)


def multiopen(pool: "multiprocessing.pool.Pool",
              func: Callable,
              l: IO,
              semaphore: Optional["multiprocessing.synchronize.Semaphore"] = None,
              nlines: Optional[int] = None,
              unordered: bool = True,
              return_num: bool = False,
              start: int = 0,
              extra: Optional[Any] = None,
              interval: Optional[int] = None,
              bar: bool = True,
              desc: Optional[str] = None,
              unit: str = "it",
              total: Optional[int] = None) -> Iterable:
    """Returns an interated object for process a file with multiple processors.

    Parameters
    ----------
    pool: multiprocessing.Pool
        The pool for multiprocessing.
    func: function
        The function to process lines.
    l: File object
        The file object.
    semaphore: multiprocessing.Semaphore, optional, default: None
        The semaphore to acquire. If None (default), the object will be passed
        without control.
    nlines: int, optional, default: None
        The number of lines to pass to the function each time. If None (default),
        only one line will be passed to the function.
    unordered: bool, optional, default: True
        Whether the process can be unordered.
    return_num: bool, optional, default: False
        If True, adds a counter to an iterable.
    start: int, optional, default: 0
        The start number of the counter.
    extra: object, optional, default: None
        The extra object passed to the item.
    interval: int, optional, default: None
        The interval of items that will be passed to the function. For example,
        if set to 10, a item will be passed once every 10 items and others will
        be dropped.
    bar: bool, optional, default: True
        If True, show a tqdm bar for the iteration.
    desc: str, optional, default: None
        The description of the iteration shown in the bar.
    unit: str, optional, default: it
        The unit of the iteration shown in the bar.
    total: int, optional, default: None
        The total number of the iteration shown in the bar.
    
    Returns
    -------
    object
        An object that can be iterated.
    """
    obj = l
    if nlines:
        obj = itertools.zip_longest(*[obj] * nlines)
    if interval:
        obj = itertools.islice(obj, 0, None, interval)
    if return_num:
        obj = enumerate(obj, start)
    if semaphore:
        obj = produce(semaphore, obj, extra)
    if unordered:
        obj = pool.imap_unordered(func, obj, 100)
    else:
        obj = pool.imap(func, obj, 100)
    if bar:
        obj = tqdm(obj, desc=desc, unit=unit, total=total)
    return obj


class SCOUROPTIONS:
    """Scour (SVG optimization) options."""
    strip_xml_prolog = True
    remove_titles = True
    remove_descriptions = True
    remove_metadata = True
    remove_descriptive_elements = True
    strip_comments = True
    enable_viewboxing = True
    strip_xml_space_attribute = True
    strip_ids = True
    shorten_ids = True
    newlines = False


class SharedRNGData:
    """Share ReacNetGenerator data with a class of the submodule.

    Parameters
    ----------
    rng: reacnetgenerator.ReacNetGenerator
        The centered ReacNetGenerator class.
    usedRNGKeys: list of strs
        Keys that needs to pass from ReacNetGenerator class to the submodule.
    returnedRNGKeys: list of strs
        Keys that needs to pass from the submodule to ReacNetGenerator class.
    extraNoneKeys: list of strs, optional, default: None
        Set keys to None, which will be used in the submodule.
    """
    def __init__(self,
                rng: "reacnetgenerator.ReacNetGenerator",
                usedRNGKeys: List[str],
                returnedRNGKeys: List[str],
                extraNoneKeys: Optional[List[str]] = None) -> None:
        self.rng = rng
        self.returnedRNGKeys = returnedRNGKeys
        for key in usedRNGKeys:
            setattr(self, key, getattr(self.rng, key))
        for key in returnedRNGKeys:
            setattr(self, key, None)
        if extraNoneKeys is not None:
            for key in extraNoneKeys:
                setattr(self, key, None)

    def returnkeys(self) -> None:
        """Return back keys to ReacNetGenerator class."""
        for key in self.returnedRNGKeys:
            setattr(self.rng, key, getattr(self, key))


def checksha256(filename: str, sha256_check: Union[str, List[str]]):
    """Check sha256 of a file is correct.
    
    Parameters
    ----------
    filename: str
        The filename.
    sha256_check: str or list of strs
        The sha256 to be checked.
    
    Returns
    -------
    bool
        Indicate whether sha256 is correct.
    """
    if not os.path.isfile(filename):
        return
    h = hashlib.sha256()
    b = bytearray(128*1024)
    mv = memoryview(b)
    with open(filename, 'rb', buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    sha256 = h.hexdigest()
    logger.info(f"SHA256 of {filename}: {sha256}")
    if sha256 in must_be_list(sha256_check):
        return True
    logger.warning("SHA256 is not correct.")
    logger.warning(open(filename).read())
    return False


async def download_file(urls: Union[str, List[str]],
                        pathfilename: str,
                        sha256: str) -> str:
    """Download files from remote urls if not exists.
    
    Parameters
    ----------
    urls: str or list of strs
        The url(s) that is available to download.
    pathfilename: str
        The downloading path of the file.
    sha256: str
        Sha256 of the file. If not None and match the file, the download will be skiped.
    
    Returns
    -------
    pathfilename: str
        The downloading path of the file.
    """
    s = requests.Session()
    s.mount('http://', HTTPAdapter(max_retries=3))
    s.mount('https://', HTTPAdapter(max_retries=3))
    # download if not exists
    if os.path.isfile(pathfilename) and (sha256 is None or checksha256(pathfilename, sha256)):
        return pathfilename

    # from https://stackoverflow.com/questions/16694907
    for url in must_be_list(urls):
        logger.info(f"Try to download {pathfilename} from {url}")
        with s.get(url, stream=True) as r, open(pathfilename, 'wb') as f:
            try:
                shutil.copyfileobj(r.raw, f)
                break
            except requests.exceptions.RequestException as e:
                logger.warning(f"Request {pathfilename} Error.", exc_info=e)
    else:
        raise RuntimeError(f"Cannot download {pathfilename}.")

    return pathfilename


async def gather_download_files(urls: List[dict]) -> None:
    """Asynchronously download files from remote urls if not exists.
    
    See download_multifiles function for details.
    
    See Also
    --------
    download_multifiles
    """
    await asyncio.gather(*[download_file(jdata["url"], jdata["fn"], jdata.get("sha256", None)) for jdata in urls])


def download_multifiles(urls: List[dict]) -> None:
    """Download multiple files from dicts.

    Parameters
    ----------
    urls: list of dicts
        The information of download files. Each dict should contain the following key:
            - url: str or list of strs
                The url(s) that is available to download.
            - pathfilename: str
                The downloading path of the file.
            - sha256: str, optional, default: None
                Sha256 of the file. If not None and match the file, the download will be skiped.
    """
    asyncio.run(gather_download_files(urls))


def run_mp(nproc: int, **arg: Any) -> Iterable[Any]:
    """Process a file with multiple processors.

    Parameters
    ----------
    nproc: int
        The number of processors to be used.
    **kwargs : dict, optional
        Other parameters can be found in the `multiopen` method.

    Yields
    ------
    object
        The yielded object from the `multiopen` method.

    See Also
    --------
    multiopen
    """
    pool = Pool(nproc, maxtasksperchild=1000)
    semaphore = Semaphore(nproc*150)
    try:
        results = multiopen(pool=pool, semaphore=semaphore, **arg)
        for item in results:
            yield item
            semaphore.release()
    except:
        logger.exception("run_mp failed")
        pool.terminate()
        raise
    else:
        pool.close()
    finally:
        pool.join()


def must_be_list(obj: Union[Any, List[Any]]) -> List[Any]:
    """Convert a object to a list if the object is not a list.
    
    Parameters
    ----------
    obj: Object
        The object to convert.

    Returns
    -------
    obj: list
        If the input object is not a list, returns a list that only contains that
        object. Otherwise, returns that object.
    """
    if isinstance(obj, list):
        return obj
    return [obj]



"""Welcome to install ReacNetGenerator.

Just use `pip install .` to install.
"""
import os
from pathlib import Path

import yaml
from distutils import log
from distutils.file_util import copy_file
from setuptools import setup, Extension
import setuptools.command.build_ext
try:
    from nodejs.node import call as node_call
except ModuleNotFoundError:
    log.info("nodejs-bin is not installed, try system node")
    import subprocess
    def node_call(args, **kwargs):
        return subprocess.call(['node', *args], **kwargs)


class BuildExtCommand(setuptools.command.build_ext.build_ext):

    def run(self):
        assert __doc__ is not None
        log.info(__doc__)
        log.info('Prepare JavaScript files with webpack...')
        this_directory = Path(__file__).parent
        webpack_dir = this_directory / "reacnetgenerator" / "static" / "webpack"
        with open(webpack_dir / ".yarnrc.yml") as f:
            yarn_path = str(Path(yaml.load(f, Loader=yaml.Loader)["yarnPath"]))
        node_call([yarn_path], cwd=webpack_dir)
        node_call([yarn_path, "start"], cwd=webpack_dir)
        try:
            assert (webpack_dir / "bundle.html").exists()
        except AssertionError:
            raise RuntimeError(
                "Fail to build bundle.html with Yarn, please retry.")
        # copy files
        try:
            os.makedirs(os.path.join(self.build_lib, 'reacnetgenerator', 'static', 'webpack'))
        except OSError:
            pass
        copy_file(
            os.path.join(this_directory, 'reacnetgenerator', 'static', 'webpack', 'bundle.html'),
            os.path.join(self.build_lib, 'reacnetgenerator', 'static', 'webpack', 'bundle.html'),
            verbose=self.verbose,
            dry_run=self.dry_run
        )
        # Add numpy headers to include_dirs
        import numpy as np
        self.include_dirs.append(np.get_include())
        setuptools.command.build_ext.build_ext.run(self)

if __name__ == '__main__':
    define_macros = []
    if os.environ.get("DEBUG", 0):
         define_macros.extend((('CYTHON_TRACE', '1'), ('CYTHON_TRACE_NOGIL', '1')))

    ext_modules = [
        Extension("reacnetgenerator.dps", sources=[
            "reacnetgenerator/dps.pyx", "reacnetgenerator/c_stack.cpp"],
            language="c++", define_macros=define_macros,
        ),
        Extension("reacnetgenerator.utils_np", sources=[
            "reacnetgenerator/utils_np.pyx"],
            language="c", define_macros=define_macros,
        ),
    ]

    setup(
          ext_modules=ext_modules,
          cmdclass={
              "build_ext": BuildExtCommand,
          },
          )



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# cython: language_level=3
# cython: linetrace=True
"""The main module of ReacNetGenerator."""
doc_run = """ReacNetGenerator: an automatic reaction network generator for reactive
molecular dynamics simulation.

Please cite: ReacNetGenerator: an automatic reaction network generator
for reactive molecular dynamic simulations, Phys. Chem. Chem. Phys.,
2020, 22 (2): 683-691, doi: 10.1039/C9CP05091D

Jinzhe Zeng (jinzhe.zeng@rutgers.edu), Tong Zhu (tzhu@lps.ecnu.edu.cn)

==================
Features
==================
* Processing of MD trajectory containing atomic coordinates or bond orders
* Hidden Markov Model (HMM) based noise filtering
* Isomers identifying accoarding to SMILES
* Generation of reaction network for visualization using force-directed
  algorithm
* Parallel computing

==================
Simple example
==================
ReacNetGenerator can process any kind of trajectory files containing 
atomic coordinates, e.g. a LAMMPS dump file prepared by running “dump 1
all custom 100 dump.reaxc id type x y z” in LAMMPS:
$ reacnetgenerator --type dump -i dump.reaxc -a C H O --nohmm
where C, H, and O are atomic names in the input file. Analysis report
will be generated automatically.

Also, ReacNetGenerator can process files containing bond information, 
e.g. LAMMPS bond file:
$ reacnetgenerator --type bond -i bonds.reaxc -a C H O --nohmm

You can running the following script for help:
$ reacnetgenerator -h
"""


import gc
import os
import time
import itertools
from enum import Enum
from multiprocessing import cpu_count
from typing import Any, List, Tuple, Union

import numpy as np

from . import __version__, __date__
from ._logging import logger
from ._detect import _Detect
from ._draw import _DrawNetwork
from ._mergeiso import _mergeISO 
from ._hmmfilter import _HMMFilter
from ._matrix import _GenerateMatrix
from ._path import _CollectPaths
from ._reachtml import _HTMLResult
from ._download import DownloadData
from .utils import must_be_list


class ReacNetGenerator:
    """Use ReacNetGenerator for trajectory analysis.
    
    Parameters
    ----------
    inputfiletype: str
        The type of the input file. The following type is allowed:
            - dump: LAMMPS dump file, which can be outputed by
              using `dump 1 all custom 100 dump.reaxc id type x y z`. See
              https://lammps.sandia.gov/doc/dump.html for details.
            - bond: LAMMPS ReaxFF bond file. See https://lammps.sandia.gov/doc/fix_reaxc_bonds.html
              for details.
            - xyz: XYZ file, which can also be outputed by LAMMPS using dump.
    inputfilename: str or list of strs
        The filename(s) of the input file, which can be either relative path or absolute path. If
        it is a list, the files will be read in order.
    atomname: tuple of strs
        The list of the atomic names in the input file, such as `('C', 'H', 'O')`. It should match
        the order of that in the input file.
    runHMM: bool, optional, default: True
        Process trajectory with Hidden Markov Model (HMM) or not. If the user find too many species
        are filtered, they can turn off this option.
    miso: int, optional, default: 0
        Merge the isomers and the highest frequency is used as the representative. 0, off 
        two available levels:
        1, merge the isomers with same atoms and same bond-network but different bond levels;
        2, merge the isomers with same atoms with different bond-network.
    pbc: bool, optional, default: True
        Use periodic boundary conditions (PBC) or not.
    cell: (3,3) array_like or (3,) array_like or (9,) array_like, optional, default: None
        The cell (box size) of the system. If None (default), the cell will be read from the input
        file. If the input file doesn't have cell information, this parameter will be necessary.
    nproc: int, optional, default: None
        The number of processors used for analysis. If None (default), the program will try to use
        all processors.
    selectatoms: str, optional, default: None
        Select an element from the atomic names, such as `C`, and only show species with this
        element in the reaction network. If None (default), the network will show all elements.
    split: int, optional, default: None
        Split number for the time axis. For example, if set to 10, the whole trajectroy will
        be divided into 10 parts and reactions of each part will be shown.
    a: (2,2) array_like, optional, default: [[0.999, 0.001], [0.001, 0.009]]
        Transition matrix A of HMM parameters. It is recommended for users to choose their own
        parameters. See the paper for details.
    b: (2,2) array_like, optional, default: [[0.6, 0.4], [0.4, 0.6]]
        Emission matrix B of HMM parameters. It is recommended for users to choose their own
        parameters. See the paper for details.
    
    Examples
    --------
    >>> from reacnetgenerator import ReacNetGenerator
    >>> rng=ReacNetGenerator(inputfiletype="dump", inputfilename="dump.ch4", atomname=['C', 'H', 'O'])
    >>> rng.runanddraw()
    """

    urls: dict
    moleculetempfilename: str
    moleculetemp2filename: str
    originfilename: str
    hmmfilename: str
    resultfilename: str

    def __init__(self, **kwargs: Any) -> None:
        """Init ReacNetGenerator."""
        logger.info(doc_run)
        logger.info(
            f"Version: {__version__}  Creation date: {__date__}")
        
        # process kwargs
        necessary_key = ['inputfiletype', 'inputfilename', 'atomname']
        default_value = {"node_color": [78/256, 196/256, 238/256], "p": [0.5, 0.5],
                            "a": [[0.999, 0.001], [0.001, 0.999]], "b": [[0.6, 0.4], [0.4, 0.6]],
                            "speciesfilter": [], "start_color": [0, 0, 1], "end_color": [1, 0, 0],
                            "nproc": cpu_count(), "pos": {}, "pbc": True, "split": 1, "n_searchspecies": 2,
                            "node_size": 200, "font_size": 6, "widthcoefficient": 1, "maxspecies": 20, "stepinterval": 1,
                            "nolabel": False,  "printfiltersignal": False, "showid": True, "runHMM": True, "SMILES": True,
                            "miso": 0,
                            "getoriginfile": False, "needprintspecies": True, "urls": [],
                            "matrix_size":100,
                            }
        none_key = ['selectatoms', 'species', 'pos', 'k', 'speciescenter', 'cell']
        accept_keys = ['atomtype', 'step', 'hmmit', 'timestep', 'steplinenum', 'N',
            'temp1it', 'originfilename', 'hmmfilename', 'moleculetempfilename', 'moleculetemp2filename',
            'allmoleculeroute', 'splitmoleculeroute']
        nparray_key = ["atomname", "p", "a", "b",
                        "node_color", "start_color", "end_color"]
        file_key = {"moleculefilename": "moname", "atomroutefilename": "route", "reactionfilename": "reaction",
                    "tablefilename": "table", "imagefilename": "svg", "speciesfilename": "species", "resultfilename": "html",
                    "jsonfilename": "json", "reactionabcdfilename": "reactionabcd"}
        assert set(necessary_key).issubset(
            set(kwargs)), "Must give neccessary key: %s" % ", ".join(necessary_key)
        assert set(kwargs).issubset(
            set(necessary_key) | set(default_value) | set(none_key) | set(file_key)), "Unsupported key"
        kwargs["inputfilename"] = must_be_list(kwargs["inputfilename"])
        for kk in default_value:
            kwargs.setdefault(kk, default_value[kk])
            if kwargs[kk] is None:
                kwargs[kk] = default_value[kk]
        for kk in itertools.chain(none_key, accept_keys):
            kwargs.setdefault(kk, None)
        for kk in file_key:
            kwargs.setdefault(
                kk, f"{kwargs['inputfilename'][0]}.{file_key[kk]}")
        for kk in nparray_key:
            kwargs[kk] = np.array(kwargs[kk])
        if not kwargs["runHMM"]:
            kwargs["getoriginfile"] = True
        if kwargs["selectatoms"] is None:
            kwargs["selectatoms"] = kwargs["atomname"]
        self.__dict__.update(kwargs)
        if self.cell is not None:
            if len(self.cell) == 9:
                self.cell = np.array(self.cell).reshape((3,3))
            elif len(self.cell) == 3:
                self.cell = np.diag(self.cell)
            else:
                raise RuntimeError("cell must be (3,3) array_like or (3,) array_like or (9,) array_like")

    def runanddraw(self,
                   run: bool = True,
                   draw: bool = True,
                   report: bool = True) -> None:
        """Analyze the trajectory from MD simulation.
        
        Parameters
        ----------
        run: bool, optional, default: True
            Process the trajectory or not, including DOWNLOAD, DETECT, HMM, PATH, and MATRIX steps.
        draw: bool, optional, default: True
            Draw the reaction network or not, i.e. NETWORK step.
        report: bool, optional, default: True
            Generate the analysis report, i.e. NETWORK step.
        """
        processthing = []
        if run:
            if self.urls:
                processthing.append(self.Status.DOWNLOAD)
            processthing.extend((
                self.Status.DETECT,
                self.Status.MISO,
                self.Status.HMM,
                self.Status.PATH,
                self.Status.MATRIX,
            ))
        if draw:
            processthing.append(self.Status.NETWORK)
        if report:
            processthing.append(self.Status.REPORT)
        self._process(processthing)

    def run(self) -> None:
        """Process MD trajectory, including DOWNLOAD, DETECT, HMM, PATH, and MATRIX steps.
        """
        processthing = []
        if self.urls:
            processthing.append(self.Status.DOWNLOAD)
        processthing.extend((
            self.Status.DETECT,
            self.Status.MISO,
            self.Status.HMM,
            self.Status.PATH,
            self.Status.MATRIX,
        ))
        self._process(processthing)

    def draw(self) -> None:
        """Draw the reaction network, i.e. NETWORK step.
        """
        self._process((self.Status.NETWORK,))

    def report(self) -> None:
        """Generate the analysis report, i.e. REPORT step.
        """
        self._process((self.Status.REPORT,))

    class Status(Enum):
        """ReacNetGenerator status.

        The ReacNetGenerator consists of several modules and algorithms to
        process the information from the given trajectory, including:

        - DOWNLOAD: Download trajectory from urls
        - DETECT: Read bond information and detect molecules
        - HMM: HMM filter
        - MISO: Merge isomers
        - PATH: Indentify isomers and collect reaction paths
        - MATRIX: Reaction matrix generation
        - NETWORK: Draw reaction network
        - REPORT: Generate analysis report
        """

        INIT = "Init"
        DETECT = "Read bond information and detect molecules"
        MISO = "Merge isomers"
        HMM = "HMM filter"
        PATH = "Indentify isomers and collect reaction paths"
        MATRIX = "Reaction matrix generation"
        NETWORK = "Draw reaction network"
        REPORT = "Generate analysis report"
        DOWNLOAD = "Download trajectory"

        def __str__(self):
            """Return describtion of the status."""
            return self.value

    def _process(self, steps: Union[List[Status], Tuple[Status, ...]]) -> None:
        """Process steps in order.

        Parameters
        ----------
        steps: tuple of ReacNetGenerator.Status
            The process that needs to be processed.
        """
        timearray = [time.perf_counter()]
        for i, runstep in enumerate(steps, 1):
            if runstep == self.Status.DETECT:
                _Detect.gettype(self).detect()
            elif runstep == self.Status.MISO:  
                _mergeISO(self).merge()
            elif runstep == self.Status.HMM:
                _HMMFilter(self).filter()
            elif runstep == self.Status.PATH:
                _CollectPaths.getstype(self).collect()
            elif runstep == self.Status.MATRIX:
                _GenerateMatrix(self).generate()
            elif runstep == self.Status.NETWORK:
                _DrawNetwork(self).draw()
            elif runstep == self.Status.REPORT:
                _HTMLResult(self).report()
            elif runstep == self.Status.DOWNLOAD:
                DownloadData(self).download_files()
            # garbage collect
            gc.collect()
            timearray.append(time.perf_counter())
            logger.info(
                f"Step {i}: Done! Time consumed (s): {timearray[-1]-timearray[-2]:.3f} ({runstep})")

        # delete tempfile
        for tempfilename in (
                self.moleculetempfilename, self.moleculetemp2filename, self.originfilename, self.hmmfilename):
            if tempfilename is not None:
                try:
                    os.remove(tempfilename)
                except OSError:
                    pass
        # Summary
        logger.info("====== Summary ======")
        logger.info(f"Total time(s): {timearray[-1]-timearray[0]:.3f} s")

# Please import ReacNetGenerator class from reacnetgenerator instead of reacnetgenerator.reacnetgen
__all__ = []
